#include <windows.h>

#include <assert.h>

#include <caml/mlvalues.h>
#include <caml/callback.h>
#include <caml/memory.h>
#include <caml/alloc.h>
#include <caml/fail.h>
#include <drawstuff/drawstuff.h>

#ifdef ARCH_ALIGN_DOUBLE
#error "This architecture is not supported."
#endif

static value caml_start;
static value caml_step;
static value caml_command;
static value caml_stop;

static value caml_exception;

#define CHECK_DOUBLE_ARRAY_SIZE(param, size) \
	if(Wosize_val(param) != Double_wosize * size) { \
		caml_invalid_argument("Array size of " #param " is not " #size "."); /* noreturn */ \
	} else

#define CHECK_DOUBLE_ARRAY_SIZE2(param, size1, size2) \
	if(Wosize_val(param) != Double_wosize * size1 && Wosize_val(param) != Double_wosize * size2) { \
		caml_invalid_argument("Array size of " #param " is neigher " #size1 " nor " #size2 "."); /* noreturn */ \
	} else

/* for callback */

void mystart(void)
{
	CAMLparam0();
	CAMLlocal1(v);
	v = caml_callback_exn(caml_start, Val_unit);
	if(Is_exception_result(v)) {
		dsStop();
		caml_register_global_root(&caml_exception);
		caml_exception = Extract_exception(v);
	}
	CAMLreturn0;
}

void mystep(int pause)
{
	CAMLparam0();
	CAMLlocal1(v);
	v = caml_callback_exn(caml_step, Val_int(pause));
	if(Is_exception_result(v)) {
		dsStop();
		caml_register_global_root(&caml_exception);
		caml_exception = Extract_exception(v);
	}
	CAMLreturn0;
}

void mycommand(int cmd)
{
	CAMLparam0();
	CAMLlocal1(v);
	v = caml_callback_exn(caml_command, Val_int(cmd));
	if(Is_exception_result(v)) {
		dsStop();
		caml_register_global_root(&caml_exception);
		caml_exception = Extract_exception(v);
	}
	CAMLreturn0;
}

void mystop(void)
{
	CAMLparam0();
	CAMLlocal1(v);
	v = caml_callback_exn(caml_stop, Val_unit);
	if(Is_exception_result(v)) {
		// no need to call dsStop(), because already stopped.
		caml_register_global_root(&caml_exception);
		caml_exception = Extract_exception(v);
	}
	CAMLreturn0;
}

static dsFunctions ds_func = {
	1,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};

static char** convert_argv(value target)
{
	assert(Is_block(target));
	assert(Tag_val(target) == 0);
	size_t num = Wosize_val(target) + 1;
	char **result =(char**)malloc(sizeof(char*) * num);
	size_t i;
	for(i = 0; i < num; ++i) {
		result[i] = String_val(Field(target, i));
	}
	result[i] = NULL;
	return result;
}

static void setup(value* caml_func_to, value caml_func_from, void (**cfunc_to)(void), void (*cfunc_from)(void))
{

	if(Is_block(caml_func_from)) { // assumes Some or None
		assert(Tag_val(caml_func_from) == 0);
		assert(Wosize_val(caml_func_from) == 1);

		*cfunc_to = cfunc_from;
		caml_register_global_root(caml_func_to);
		*caml_func_to = Field(caml_func_from, 0);
	} else {
		*cfunc_to = NULL;
	}
}

/* stub functions */

CAMLprim value stub_dsSimulationLoop_native(value start, value step, value command, value stop, value path, value argv, value width, value height)
{
	CAMLparam5(start, step, command, stop, path);
	CAMLxparam3(argv, width, height);
	CAMLlocal1(exception);

	char **cargv = convert_argv(argv);

	setup(&caml_start, start, &ds_func.start, mystart);
	setup(&caml_step, step, (void(**)(void))&ds_func.step, (void(*)(void))mystep);
	setup(&caml_command, command, (void(**)(void))&ds_func.command, (void(*)(void))mycommand);
	setup(&caml_stop, stop, &ds_func.stop, mystop);

	if(Is_block(path)) { // assumes Some or None
		assert(Tag_val(path) == 0);
		assert(Wosize_val(path) == 1);

		ds_func.version = 2;
		ds_func.path_to_textures = String_val(Field(path, 0));
	} else {
		ds_func.version = 1;
		ds_func.path_to_textures = NULL;
	}

	caml_exception = Val_int(0);
	dsSimulationLoop(Wosize_val(argv), cargv, Int_val(width), Int_val(height), &ds_func);

	if(ds_func.start)   caml_remove_global_root(&caml_start);
	if(ds_func.step)    caml_remove_global_root(&caml_step);
	if(ds_func.command) caml_remove_global_root(&caml_command);
	if(ds_func.stop)    caml_remove_global_root(&caml_stop);

	free(cargv);

	if(caml_exception != Val_int(0)) {
		exception = caml_exception;
		caml_remove_global_root(&caml_exception);
		caml_raise(exception); // noreturn
	}

	CAMLreturn(Val_unit);
}

CAMLprim value stub_dsSimulationLoop_byte(value *argv, int argn)
{
	stub_dsSimulationLoop_native(argv[0], argv[1], argv[2], argv[3], argv[4], argv[5], argv[6], argv[7]);
	return Val_unit;
}

CAMLprim value stub_dsSetViewpoint(value xyz, value hpr)
{
	CAMLparam2(xyz, hpr);
	float cxyz[3], chpr[3];
	size_t i;

	assert(Is_block(xyz));
	assert(Tag_val(xyz) == Double_array_tag);

	assert(Is_block(hpr));
	assert(Tag_val(hpr) == Double_array_tag);

	CHECK_DOUBLE_ARRAY_SIZE(xyz, 3);
	CHECK_DOUBLE_ARRAY_SIZE(hpr, 3);

	for(i = 0; i < 3; ++i) {
		cxyz[i] = (float)Double_field(xyz, i);
		chpr[i] = (float)Double_field(hpr, i);
	}

	dsSetViewpoint(cxyz, chpr);

	CAMLreturn(Val_unit);
}

CAMLprim value stub_dsGetViewpoint(value xyz, value hpr)
{
	CAMLparam2(xyz, hpr);

	float cxyz[3];
	float chpr[3];
	size_t i;

	assert(Is_block(xyz));
	assert(Tag_val(xyz) == Double_array_tag);

	CHECK_DOUBLE_ARRAY_SIZE(xyz, 3);

	assert(Is_block(hpr));
	assert(Tag_val(hpr) == Double_array_tag);

	CHECK_DOUBLE_ARRAY_SIZE(hpr, 3);

	dsGetViewpoint(cxyz, chpr);

	for(i = 0; i < 3; ++i) {
		Store_double_field(xyz, i, cxyz[i]);
		Store_double_field(hpr, i, chpr[i]);
	}

	CAMLreturn(Val_unit);
}

CAMLprim value stub_dsStop(value unit)
{
	CAMLparam1(unit);
	dsStop();
	CAMLreturn(Val_unit);
}

CAMLprim value stub_dsElapsedTime(value unit)
{
	CAMLparam1(unit);
	CAMLlocal1(result);
	result = caml_alloc(Double_tag, Double_wosize);
	Store_double_val(result, dsElapsedTime());
	CAMLreturn(result);
}

CAMLprim value stub_dsSetTexture(value number)
{
	CAMLparam1(number);
	dsSetTexture(number);
	CAMLreturn(Val_unit);
}

CAMLprim value stub_dsSetColor(value red, value green, value blue)
{
	CAMLparam3(red, green, blue);
	dsSetColor(Double_val(red), Double_val(green), Double_val(blue));
	CAMLreturn(Val_unit);
}

CAMLprim value stub_dsSetColorAlpha(value red, value green, value blue, value alpha)
{
	CAMLparam4(red, green, blue, alpha);
	dsSetColorAlpha(Double_val(red), Double_val(green), Double_val(blue), Double_val(alpha));
	CAMLreturn(Val_unit);
}

CAMLprim value stub_dsDrawBox(value p, value R, value sides)
{
	CAMLparam3(p, R, sides);

	CHECK_DOUBLE_ARRAY_SIZE2(p, 3, 4);
	CHECK_DOUBLE_ARRAY_SIZE(R, 12);
	CHECK_DOUBLE_ARRAY_SIZE(sides, 3);

	dsDrawBoxD((double*)p, (double*)R, (double*)sides);
	CAMLreturn(Val_unit);
}

CAMLprim value stub_dsDrawSphere(value p, value R, value radius)
{
	CAMLparam3(p, R, radius);

	CHECK_DOUBLE_ARRAY_SIZE2(p, 3, 4);
	CHECK_DOUBLE_ARRAY_SIZE(R, 12);

	dsDrawSphereD((double*)p, (double*)R, Double_val(radius));
	CAMLreturn(Val_unit);
}

CAMLprim value stub_dsDrawTriangle_native(value p, value R, value v0, value v1, value v2, value solid)
{
	CAMLparam5(p, R, v0, v1, v2);
	CAMLxparam1(solid);

	CHECK_DOUBLE_ARRAY_SIZE2(p, 3, 4);
	CHECK_DOUBLE_ARRAY_SIZE(R, 12);
	CHECK_DOUBLE_ARRAY_SIZE(v0, 3);
	CHECK_DOUBLE_ARRAY_SIZE(v1, 3);
	CHECK_DOUBLE_ARRAY_SIZE(v2, 3);

	dsDrawTriangleD((double*)p, (double*)R, (double*)v0, (double*)v1, (double*)v2, Int_val(solid));
	CAMLreturn(Val_unit);
}

CAMLprim value stub_dsDrawTriangle_byte(value *argv, int argn)
{
	stub_dsDrawTriangle_native(argv[0], argv[1], argv[2], argv[3], argv[4], argv[5]);
	return Val_unit;
}

CAMLprim value stub_dsDrawCylinder(value p, value R, value length, value radius)
{
	CAMLparam4(p, R, length, radius);

	CHECK_DOUBLE_ARRAY_SIZE2(p, 3, 4);
	CHECK_DOUBLE_ARRAY_SIZE(R, 12);

	dsDrawCylinderD((double*)p, (double*)R, Double_val(length), Double_val(radius));
	CAMLreturn(Val_unit);
}

CAMLprim value stub_dsDrawCapsule(value p, value R, value length, value radius)
{
	CAMLparam4(p, R, length, radius);

	CHECK_DOUBLE_ARRAY_SIZE2(p, 3, 4);
	CHECK_DOUBLE_ARRAY_SIZE(R, 12);

	dsDrawCapsuleD((double*)p, (double*)R, Double_val(length), Double_val(radius));
	CAMLreturn(Val_unit);
}

CAMLprim value stub_dsDrawLine(value p1, value p2)
{
	CAMLparam2(p1, p2);

	CHECK_DOUBLE_ARRAY_SIZE2(p1, 3, 4);
	CHECK_DOUBLE_ARRAY_SIZE2(p2, 3, 4);

	dsDrawLineD((double*)p1, (double*)p2);
	CAMLreturn(Val_unit);
}

static double* convert_planes(value planes)
{
	size_t num = 0;
	value temp = planes;
	double *result;
	while(Is_block(temp)) {
		++num;
		temp = Field(temp, 1);
	}
	result =(double*)malloc(sizeof(double) * num * 4);
	temp = planes;
	size_t i;
	for(i = 0; i < num; ++i) {
		result[i*4+0] = Double_field(Field(temp, 0), 0);
		result[i*4+1] = Double_field(Field(temp, 0), 1);
		result[i*4+2] = Double_field(Field(temp, 0), 2);
		temp = Field(temp, 1);
	}
	return result;
}

static double* convert_points(value points)
{
	size_t num = 0;
	value temp = points;
	double *result;
	while(Is_block(temp)) {
		++num;
		temp = Field(temp, 1);
	}
	result =(double*)malloc(sizeof(double) * num * 3);
	temp = points;
	size_t i;
	for(i = 0; i < num; ++i) {
		result[i*3+0] = Double_field(Field(temp, 0), 0);
		result[i*3+1] = Double_field(Field(temp, 0), 1);
		result[i*3+2] = Double_field(Field(temp, 0), 2);
		temp = Field(temp, 1);
	}
	return result;
}

static unsigned int* convert_polygons(value polygons)
{
	size_t num = 0, num2 = 0;
	size_t i, j, k;
	value temp = polygons, temp2;
	unsigned int *result;
	while(Is_block(temp)) {
		++num; ++num2;
		temp2 = Field(temp, 0);
		while(Is_block(temp2)) {
			++num2;
			temp2 = Field(temp2, 1);
		}
		temp = Field(temp, 1);
	}
	result =(unsigned int*)malloc(sizeof(unsigned int) * num2);
	temp = polygons;
	for(i = 0, j = 0; i < num; ++i) {
		k = j++;
		num2 = 0;
		temp2 = Field(temp, 0);
		while(Is_block(temp2)) {
			result[j++] = Field(temp2, 0);
			++num2;
			temp2 = Field(temp2, 1);
		}
		result[k] = num2;
		temp = Field(temp, 1);
	}
	return result;
}

CAMLprim value stub_dsDrawConvex(value p, value R, value planes, value points, value polygons)
{
	CAMLparam5(p, R, planes, points, polygons);
	double *cplanes, *cpoints;
	unsigned int *cpolygons;

	CHECK_DOUBLE_ARRAY_SIZE2(p, 3, 4);
	CHECK_DOUBLE_ARRAY_SIZE(R, 12);

	cplanes = convert_planes(planes);
	cpoints = convert_points(points);
	cpolygons = convert_polygons(polygons);

	dsDrawConvexD((double*)p, (double*)R, cplanes, Wosize_val(planes), cpoints, Wosize_val(points), cpolygons);

	free(cpolygons);
	free(cpoints);
	free(cplanes);

	CAMLreturn(Val_unit);
}

CAMLprim value stub_dsSetSphereQuality(value n)
{
	CAMLparam1(n);
	dsSetSphereQuality(Int_val(n));
	CAMLreturn(Val_unit);
}

CAMLprim value stub_dsSetCapsuleQuality(value n)
{
	CAMLparam1(n);
	dsSetCapsuleQuality(Int_val(n));
	CAMLreturn(Val_unit);
}
